import time
from typing import Optional
import socket
import json
import subprocess
import sys

from rune.models.session.protocol.command import EndSessionCmd, GetSessionKeyCmd, HandshakeCmd, SessionCmd, SessionStatusCmd, StartSessionCmd
from rune.models.session.protocol.response import FailureResponse, GetKeyResponse, HandshakeResp, SessionResp, StatusResponse, SuccessResponse
from rune.models.session.status import SessionStatus
from rune.session.base import SessionManager
from rune.utils.environment import sanitized_env
from rune.exception.session import NoSessionError, WrongUserError


class DaemonSessionManager(SessionManager):
    def __init__(self, daemon_host: str, daemon_port: int) -> None:
        self.daemon_host = daemon_host
        self.daemon_port = daemon_port

    def start_session(self, user: str, session_key: str, ttl_seconds: int) -> None:
        """
        Starts a session.
        Session exists for the provided ttl.
        """
        if not self._is_daemon_started():
            self._spawn_daemon()

        self.make_request(StartSessionCmd(session_key, ttl_seconds, user))


    def end_session(self) -> None:
        """
        Ends the current session.

        raises NoSessionError if the session does not exist.
        """
        if not self._is_daemon_started():
            raise NoSessionError("No session in progress.")

        resp = self.make_request(EndSessionCmd())
        if isinstance(resp, SuccessResponse):
            return

        raise RuntimeError(str(resp.to_dict()))

    def is_session_in_progress(self) -> bool:
        """
        Returns true if there is an ongoing session

        False otherwise.
        """
        if not self._is_daemon_started():
            return False

        resp = self.make_request(SessionStatusCmd())
        if isinstance(resp, StatusResponse):
            return resp.remaining_ttl != -1

        return False


    def get_default_key(self, user: str) -> Optional[str]:
        """
        Retrieves the default key for this session.
        Returns None if the key is not set.

        Raises NoSessionError if the session does not exist.
        Raises WrongUserError if the user is not the same as the one that started the session.
        """
        if not self._is_daemon_started():
            raise NoSessionError("No session in progress.")

        resp = self.make_request(GetSessionKeyCmd(user))
        if isinstance(resp, FailureResponse):
            raise WrongUserError(resp.message)

        if isinstance(resp, GetKeyResponse):
            return resp.session_key

        return None

    def get_session_status(self) -> SessionStatus:
        """
        Retrieves the session status.

        Raises NoSessionError if the session does not exist.
        """
        try:
            resp = self.make_request(SessionStatusCmd())
            if isinstance(resp, StatusResponse):
                return SessionStatus(True, resp.remaining_ttl, resp.user)
            return SessionStatus.STARTED_UNKNOWN()
        except:
            return SessionStatus.NOT_STARTED()

    def _is_daemon_started(self, timeout: float = 1) -> bool:
        try:
            resp = self.make_request(HandshakeCmd(), timeout)
            return isinstance(resp, HandshakeResp) and resp.all_good
        except:
            return False

    def _spawn_daemon(self) -> bool:
        env = sanitized_env({
            "HOST": self.daemon_host,
            "PORT": str(self.daemon_port),
        })

        subprocess.Popen(
            [sys.executable, "-m", "rune.session.daemon"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            env=env,
            close_fds=True,
            start_new_session=True,
        )

        time.sleep(0.5)

        return self._is_daemon_started()

    def make_request(self, request: SessionCmd, timeout: float = 1) -> SessionResp:
        with socket.create_connection((self.daemon_host, self.daemon_port)) as s:
            s.sendall((json.dumps(request.to_dict()) + "\n").encode("utf-8"))

            s.settimeout(timeout)
            response = s.recv(4096)
            return SessionResp.from_dict(json.loads(response.decode("utf-8")))







import os
from typing import Dict

def sanitized_env(extra: Dict[str, str] | None = None) -> Dict[str, str]:
    env = os.environ.copy()

    PYTHON_BAD = {
        "PYTHONPATH",
        "PYTHONHOME",
        "PYTHONSTARTUP",
        "PYTHONINSPECT",
        "PYTHONDONTWRITEBYTECODE",
        "PYTHONBREAKPOINT",
        "PYTHONTRACEMALLOC",
        "PYTHONASYNCIODEBUG",
        "PYTHONWARNINGS",
    }

    LOADER_BAD = {
        "LD_PRELOAD",
        "LD_LIBRARY_PATH",
        "DYLD_INSERT_LIBRARIES",
        "DYLD_LIBRARY_PATH",
    }

    for key in PYTHON_BAD | LOADER_BAD:
        env.pop(key, None)

    for key in ("TERM", "COLUMNS", "LINES"):
        env.pop(key, None)

    if extra:
        for k, v in extra.items():
            env[k] = str(v)

    return env




from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Dict, ClassVar, Type, Self


class SessionCmd(ABC):
    START_SESSION: ClassVar[str] = "start"
    END_SESSION: ClassVar[str] = "end"
    GET_SESSION_KEY: ClassVar[str] = "get"
    SESSION_STATUS: ClassVar[str] = "status"
    HANDSHAKE: ClassVar[str] = "handshake"

    _registry: ClassVar[Dict[str, Type["SessionCmd"]]] = {}

    CMD: ClassVar[str]

    def to_dict(self) -> Dict:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, d: Dict) -> "SessionCmd":
        try:
            cmd = d["type"]
        except:
            raise ValueError(f"Got unexpected command type: {d.get("type")}") from None

        match cmd:
            case cls.START_SESSION: return StartSessionCmd.from_dict(d)
            case cls.END_SESSION: return EndSessionCmd.from_dict(d)
            case cls.GET_SESSION_KEY: return GetSessionKeyCmd.from_dict(d)
            case cls.SESSION_STATUS: return SessionStatusCmd.from_dict(d)
            case cls.HANDSHAKE: return HandshakeCmd.from_dict(d)

        raise ValueError(f"Got unexpected command type: {d.get("type")}") from None

@dataclass
class StartSessionCmd(SessionCmd):
    CMD: ClassVar[str] = SessionCmd.START_SESSION

    session_key: str
    ttl: int
    user: str

    def to_dict(self) -> Dict:
        return {
            "type": self.CMD,
            "session_key": self.session_key,
            "ttl": self.ttl,
            "user": self.user
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["session_key"], d["ttl"], d["user"])


@dataclass
class EndSessionCmd(SessionCmd):
    CMD: ClassVar[str] = SessionCmd.END_SESSION

    def to_dict(self) -> Dict:
        return {
            "type": self.CMD,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls()


@dataclass
class GetSessionKeyCmd(SessionCmd):
    CMD: ClassVar[str] = SessionCmd.GET_SESSION_KEY

    user: str

    def to_dict(self) -> Dict:
        return {
            "type": self.CMD,
            "user": self.user
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["user"])


@dataclass
class SessionStatusCmd(SessionCmd):
    CMD: ClassVar[str] = SessionCmd.SESSION_STATUS

    def to_dict(self) -> Dict:
        return {"type": self.CMD}

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls()

@dataclass
class HandshakeCmd(SessionCmd):
    CMD: ClassVar[str] = SessionCmd.HANDSHAKE

    def to_dict(self) -> Dict:
        return {"type": self.CMD}

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls()




from abc import ABC, abstractmethod
from typing import Optional

from rune.models.session.status import SessionStatus


class SessionManager(ABC):
    @abstractmethod
    def start_session(self, user: str, session_key: str, ttl_seconds: int) -> None:
        """
        Starts a session.
        Session exists for the provided ttl. (-1 means it will not close)
        """
        raise NotImplementedError()

    @abstractmethod
    def end_session(self) -> None:
        """
        Ends the current session.

        raises NoSessionError if the session does not exist.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_session_in_progress(self) -> bool:
        """
        Returns true if there is an ongoing session

        False otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_default_key(self, user: str) -> Optional[str]:
        """
        Retrieves the default key for this session.
        Returns None if the key is not set.

        Raises NoSessionError if the session does not exist.
        Raises WrongUserError if the user is not the same as the one that started the session.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_session_status(self) -> SessionStatus:
        """
        Retrieves the session status.

        Raises NoSessionError if the session does not exist.
        """
        raise NotImplementedError()





from typing import Annotated, Optional
from typer import Typer, Option

from rune.exception.session import NoSessionError, WrongUserError
from rune.utils.input import ensure_active_user, input_default_key
from rune.context import Context
from rune.utils import display

DEFAULT_KEY_HELP = (
    "Default encryption key for this session.\n"
    "If omitted, you will be prompted securely.\n"
    "The key is kept in memory only and never written to disk."
)

SESSION_TTL_HELP = (
    "Session time-to-live in seconds.\n"
    "If omitted, the configured default TTL is used.\n"
    "Use -1 to create a session that does not expire."
)

def setup() -> Typer:

    session_app = Typer(name="session", help="Manage rune session. Run `rune session -h` for more help.")
    @session_app.command(name="start")
    def start(
        _key: Annotated[
            Optional[str],
            Option("--session-key", "-k", help=DEFAULT_KEY_HELP)
        ] = None,
        _ttl: Annotated[
            Optional[int],
            Option("--ttl", help=SESSION_TTL_HELP)
        ] = None
    ):
        """
        Start a local rune session.

        A session holds a default encryption key in memory and makes it
        available to rune commands for encrypting and decrypting secrets.

        Sessions are local-only and scoped to the active username.
        """
        key = _key or input_default_key()
        username = ensure_active_user()
        context = Context.get()
        sessionmgr = context.session_manager
        ttl = _ttl or context.settings.session.default_ttl

        sessionmgr.start_session(username, key, ttl)

        expiry_message = (
            "Session will [bold]not expire[/]."
            if ttl == -1
            else f"Session will expire in [bold]{ttl} seconds[/]."
        )
        display.success_panel(
            f"Session started for user [bold cyan]{username}[/]. {expiry_message}"
        )

    @session_app.command(name="end")
    def end():
        """
        End the active session.

        This immediately clears the default encryption key from memory
        and shuts down the session daemon.
        """
        sessionmgr = Context.get().session_manager

        try:
            sessionmgr.end_session()
            display.success_panel("Session ended.")
        except NoSessionError:
            display.failed_panel("No active session to end.")


    @session_app.command(name="status")
    def status():
        """
        Show the current session status.

        Displays whether a session is running, which user it belongs to,
        and its remaining TTL (if applicable).
        """
        sessionmgr = Context.get().session_manager

        status = sessionmgr.get_session_status()
        display.success_panel(
            f"[bold]Started:[/] {status.started}\n"
            f"[bold]TTL:[/]     {status.ttl or 'N/A'}\n"
            f"[bold]User:[/]    {status.user or 'N/A'}"
        )

    return session_app




from dataclasses import dataclass
from typing import Self


@dataclass
class SessionStatus:
    started: bool
    ttl: int | None
    user: str | None

    @classmethod
    def NOT_STARTED(cls) -> Self:
        return cls(False, None, None)

    @classmethod
    def STARTED_UNKNOWN(cls) -> Self:
        return cls(True, None, None)



from abc import ABC, abstractmethod
from typing import Optional

class Result[V](ABC):
    @abstractmethod
    def is_success(self) -> bool:
        raise NotImplementedError()
    @abstractmethod
    def is_failure(self) -> bool:
        raise NotImplementedError()
    def value(self) -> Optional[V]:
        return None
    def failure_reason(self) -> Optional[str]:
        return None

class Success[V](Result[V]):
    def __init__(self, value: V = None) -> None:
        self._value = value
    def is_success(self) -> bool: return True
    def is_failure(self) -> bool: return False
    def value(self) -> Optional[V]:
        return self._value

class Failure[V](Result[V]):
    def __init__(self, failure_reason: str) -> None:
        self._failure_reason = failure_reason
    def is_success(self) -> bool: return False
    def is_failure(self) -> bool: return True

    def failure_reason(self) -> Optional[str]:
        return self._failure_reason









from dataclasses import dataclass, field
from typing import Optional, Dict, Self

@dataclass
class SecretField:
    ciphertext: str
    nonce: Optional[str] = None
    tag: Optional[str] = None
    salt: Optional[str] = None
    algorithm: Optional[str] = None
    deleted: bool = False
    params: Dict[str, str] = field(default_factory=dict)

    version: int = 1

    def soft_delete(self) -> Self:
        self.deleted = True
        return self

    def restore(self) -> Self:
        self.deleted = False
        return self

    def to_dict(self) -> Dict:
        return {
            "ciphertext": self.ciphertext,
            "nonce": self.nonce,
            "tag": self.tag,
            "salt": self.salt,
            "algorithm": self.algorithm,
            "deleted": self.deleted,
            "params": self.params,
            "version": self.version
        }

    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        return cls(
            ciphertext=data["ciphertext"],
            nonce=data.get("nonce"),
            tag=data.get("tag"),
            salt=data.get("salt"),
            algorithm=data.get("algorithm"),
            deleted=data.get("deleted", False),
            params=data.get("params", {}),
            version=data.get("version", 1)
        )





from abc import ABC, abstractmethod
from typing import Dict, Self
from dataclasses import dataclass

class SessionSettings(ABC):
    DAEMON: str = "daemon"

    @property
    def mode(self) -> str: raise NotImplementedError()

    @property
    def default_ttl(self) -> int: raise NotImplementedError()

    @abstractmethod
    def to_dict(self) -> Dict: raise NotImplementedError()

    @classmethod
    def from_dict(cls, d: Dict | None) -> "SessionSettings":
        if not d: return cls.default()
        match d["mode"]:
            case cls.DAEMON:
                return DaemonSessionSettings.from_dict(d)
            case _:
                raise ValueError(f"Unknown encryption mode '{d["mode"]}'")

    @classmethod
    def from_mode(cls, mode: str) -> "SessionSettings":
        match mode:
            case cls.DAEMON:
                return DaemonSessionSettings.default()
            case _:
                raise ValueError(f"Unknown encryption mode '{mode}'")

    @classmethod
    def default(cls) -> "SessionSettings":
        return DaemonSessionSettings.default()



@dataclass
class DaemonSessionSettings(SessionSettings):
    port: int
    _default_ttl: int

    @property
    def mode(self) -> str: return SessionSettings.DAEMON

    @property
    def default_ttl(self) -> int: return self._default_ttl

    def to_dict(self) -> Dict:
        return {
            "mode": self.mode,
            "port": self.port,
            "default_ttl": self.default_ttl,
        }
    
    @classmethod
    def from_dict(cls, d: Dict | None) -> Self:
        if not d: return cls.default()
        return cls(
            port=d.get("port", cls.default().port),
            _default_ttl = d.get("default_ttl", cls.default().default_ttl)
        )

    @classmethod
    def default(cls) -> Self:
        return cls(
            port=5000,
            _default_ttl=300
        )





import socket
import json
import time

from rune.models.session.protocol.response import FailureResponse, HandshakeResp, SessionResp, GetKeyResponse, StatusResponse, SuccessResponse
from rune.models.session.protocol.command import EndSessionCmd, GetSessionKeyCmd, HandshakeCmd, SessionCmd, SessionStatusCmd, StartSessionCmd

class State:
    def __init__(self) -> None:
        self.start_time: float
        self.user: str | None = None
        self.session_key: str | None = None
        self.started: bool = False
        self.end_time: float | None = None
        self.force_finish: bool = False
        self.no_ttl: bool = False

    @property
    def time_remaining(self) -> float:
        if self.no_ttl:
            return -1
        if self.end_time:
            return self.end_time - time.time()
        return -1

    @property
    def is_finished(self) -> bool:
        if not self.started:
            return False
        if self.force_finish:
            return True
        if self.no_ttl:
            return False
        if self.time_remaining:
            return self.time_remaining < 0
        return False

    def start(self, user: str, session_key: str, ttl_seconds: int) -> None:
        if ttl_seconds == -1:
            self.no_ttl = True

        self.start_time = time.time()
        self.end_time = self.start_time + ttl_seconds

        self.user = user
        self.started = True
        self.session_key = session_key

    def end(self) -> None:
        self.force_finish = True

def handle_client(conn, addr, state: State):
    with conn:
        buffer = ""
        while True:
            data = conn.recv(4096)
            if not data:
                break

            buffer += data.decode("utf-8")

            while "\n" in buffer:
                line, buffer = buffer.split("\n", 1)

                try:
                    request = SessionCmd.from_dict(json.loads(line))
                    response = process_request(request, state)
                except json.JSONDecodeError:
                    response = FailureResponse("Error decoding request.")


                conn.sendall((json.dumps(response.to_dict()) + "\n").encode("utf-8"))


def process_request(request: SessionCmd, state: State) -> SessionResp:
    cmd = request.CMD

    match cmd:
        case SessionCmd.GET_SESSION_KEY if isinstance(request, GetSessionKeyCmd):
            if state.user != request.user:
                return FailureResponse("Stored Session Key was provided by a different user.")
            if state.session_key:
                return GetKeyResponse(state.session_key)
            else:
                return FailureResponse("Session key is not set")
        case SessionCmd.START_SESSION if isinstance(request, StartSessionCmd):
            if not state.started:
                state.start(request.user, request.session_key, request.ttl)
                return SuccessResponse("Session started")
            else:
                return FailureResponse("Session already in progress")
        case SessionCmd.END_SESSION if isinstance(request, EndSessionCmd):
            if state.started:
                state.end()
                return SuccessResponse("Session ended")
            else:
                return FailureResponse("No session in progress")
        case SessionCmd.SESSION_STATUS if isinstance(request, SessionStatusCmd):
            if state.time_remaining:
                return StatusResponse(int(state.time_remaining), state.user or "")
            else:
                return StatusResponse(-1, state.user or "")
        case SessionCmd.HANDSHAKE if isinstance(request, HandshakeCmd):
            return HandshakeResp(all_good=True)

    return FailureResponse(f"Unknown command type {cmd} or session format {str(type(request))}.")


def main(host: str, port: int):
    state = State()
    print(state)
    print(state.is_finished)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        s.bind((host, port))
        s.listen()

        while not state.is_finished:
            try:
                conn, addr = s.accept()
                handle_client(conn, addr, state)
            except TimeoutError:
                continue




from rune.internal.get import get_secret
from rune.utils.input import get_choice_by_idx, input_key, input_name, sanitize_name, get_session_key
from rune.utils import display

import pyperclip


def handle_get_command(user: str, _name: str | None, _key: str | None, show: bool, show_deleted: bool, use_session_key: bool):
    name = sanitize_name(_name or input_name())

    session_key = get_session_key(user) if use_session_key else None
    key = _key or session_key or input_key()

    result = get_secret(user, name, key)

    v = result.value()

    if not (result.is_success() and v):
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")
        return
    
    to_display = {k:s[0] for k, s in v.items() if show_deleted or not s[1]}

    display.panel(
        f"Stored fields for secret '[bold cyan]{name}[/]':\n" + \
        "\n".join([f"[bold cyan][{i}][/] {k}" + ("" if not show else f" - {s}") for i, (k, s) in enumerate(to_display.items(), 1)])
    )

    while True:
        choice = get_choice_by_idx("[cyan]Select field to copy[/]", list(to_display.values()))
        if not choice:
            return
        pyperclip.copy(choice)
        display.success_panel(f"[bold green]✓ Copied to clipboard[/]", title="Copied")




from dataclasses import dataclass, field
from typing import List, Self

@dataclass
class Namespace:
    ns: List[str] = field(default_factory=list)

    @property
    def to_string(self):
        return "/".join(self.ns).removeprefix("/").removesuffix("/").strip()

    @classmethod
    def from_string(cls, s: str | None) -> Self:
        if not s:
            return cls()
        else:
            parts = s.split("/")
            return cls(parts)




from .process import main
import os

if __name__ == "__main__":
    host = os.environ.get("HOST")
    port_str = os.environ.get("PORT")
    if not (host and port_str):
        print("HOST and PORT environment variables are not set. Not spawning daemon.")
        exit(2)
    try:
        port = int(port_str)
    except:
        print("Bad port. Not spawning daemon.")
        exit(2)

    main(host, port)



from pathlib import Path
from typer import Argument, Typer, Option
from typing import Annotated, Literal, Optional
import pyperclip

from rune.context import Context
from rune.models.settings.encryptionsettings import EncryptionSettings
from rune.models.settings.sessionsettings import DaemonSessionSettings
from rune.models.settings.storagesettings import FileBasedStorageSettings
from rune.utils.input import get_choice_by_idx, require
from rune.utils import display

STORAGE_MODE_HELP = (
    "Configure how rune stores encrypted secrets.\n"
    "Currently supported modes:\n"
    "  - local: store secrets in a local file on disk."
)

STORAGE_FILE_HELP = (
    "Path to the secrets file when using local storage.\n"
    "Example: ~/.rune/secrets.json"
)

ENCRYPTION_MODE_HELP = (
    "Configure the encryption algorithm used to encrypt secrets.\n"
    "Changing this does NOT re-encrypt existing secrets."
)

def setup() -> Typer:
    """
    Register configuration-related commands.
    """

    config_app = Typer(name="config", help="Manage rune configs. Run `rune config -h` for more help.")

    @config_app.command(name="storage")
    def config_storage(
        _mode: Annotated[
        Optional[Literal["local"]],
        Option("--mode", "-m", help=STORAGE_MODE_HELP),
    ] = None,
        _file: Annotated[
        Optional[str],
        Option("--file", "-f", help=STORAGE_FILE_HELP),
    ] = None,
    ):
        """
        Configure how and where secrets are stored.
        """
        context = Context.get()
        mode: str = _mode or context.settings.storage.mode

        if mode == "local":
            file = require(_file, "A file path is required when using local storage.")
            path = Path(file)
            storage_path = str(path.expanduser().absolute())

            new_settings = FileBasedStorageSettings(storage_path)
            context.settings.update(storage=new_settings)

            display.success_panel(f"Storage file set to: [bold cyan]{storage_path}[/]\n[dim]Existing secrets were not modified.[/]", title="Storage Updated")

    @config_app.command(name="encryption")
    def config_encryption(
        mode: Annotated[
        Literal["aesgcm"],
        Option("--mode", "-m", help=ENCRYPTION_MODE_HELP),
    ],
    ):
        """
        Configure the encryption algorithm used by rune.
        """
        context = Context.get()

        if mode == context.settings.encryption.mode:
            display.failed_panel(f"Encryption mode is already set to '[bold]{mode}[/]'.", title="Nothing changed")
            return

        new_settings = EncryptionSettings.from_mode(mode)
        context.settings.update(encryption=new_settings)

        display.success_panel(f"Encryption mode set to '[bold]{mode}[/]'.\n\n" +\
            "[dim]Existing secrets remain encrypted with their original settings.[/]", title="Encryption Updated")

    @config_app.command(name="session")
    def config_session(
        _port: Annotated[int, Option("--port", "-p", help="The port for the local session daemon.")],
        _default_ttl: Annotated[Optional[int], Option("--default-ttl", "-ttl", help="The default ttl for the provided default key.")] = None,
    ):
        """
        Configure the session daemon used by rune.
        """
        context = Context.get()

        new_settings = DaemonSessionSettings(
            port=_port,
            _default_ttl=_default_ttl or context.settings.session.default_ttl,
        )
        context.settings.update(session=new_settings)

        display.success_panel(f"Session daemon port set to '[bold]{_port}[/]'.", title="Encryption Updated")



    @config_app.command(name="show")
    def show_config(
        profile: Annotated[
        Optional[str],
        Argument(
            help=(
                "Profile name to display.\n"
                    "If omitted, shows the currently active configuration."
            )
        ),
    ] = None,
    ):
        """
        Display the current configuration or a specific profile.
        """
        context = Context.get()

        if not profile:
            settings_file = context.settings_manager.settings_file
            settings = context.settings.to_dict()

            display.print("[bold]Active configuration:[/]")
            display.print(f"Settings file: [cyan]{settings_file}[/]")
            display.print_dict(settings)
            return

        settings = context.settings_manager.get_profile(profile)

        if not settings:
            display.failed_panel(f"Profile '[bold cyan]{profile}[/]' does not exist.")
            return

        display.print(f"[bold]Configuration for profile '[cyan]{profile}[/]':[/]")
        display.print_dict(settings.to_dict())

    @config_app.command(name="where")
    def whereis(
        interactive: Annotated[
        bool,
        Option(
            "--interactive",
            "-i",
            help="Interactively select a file path to copy to clipboard.",
        ),
    ] = False
    ):
        """
        Show where rune stores its configuration and profile files.
        """
        context = Context.get()
        settings_file = str(context.settings_manager.settings_file.absolute())
        profiles_file = str(context.settings_manager.profiles_file.absolute())

        if not interactive:
            display.panel(f"[bold]Settings file[/]: [cyan]{settings_file}[/]\n[bold]Profiles file[/]: [cyan]{profiles_file}[/]")
            return

        choices = [settings_file, profiles_file]

        display.panel(
            f"[bold cyan][1][/] Settings file: {settings_file}\n"
                f"[bold cyan][2][/] Profiles file: {profiles_file}",
            title="File Locations",
        )

        choice = get_choice_by_idx("Copy file path", choices)
        if choice:
            pyperclip.copy(choice)
            display.print("[green]Path copied to clipboard.[/]")

    return config_app




from __future__ import annotations

from abc import ABC
from typing import Dict, ClassVar, Self


class SessionResp(ABC):
    STATUS: ClassVar[str] = "status"
    GET_KEY: ClassVar[str] = "get_key"
    SUCCESS: ClassVar[str] = "success"
    FAILURE: ClassVar[str] = "failure"
    HANDSHAKE: ClassVar[str] = "handshake"

    RESP: ClassVar[str]

    def to_dict(self) -> Dict:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, d: Dict) -> "SessionResp":
        try:
            resp_type = d["type"]
        except:
            raise ValueError(f"Got unexpected response type: {d.get("type")}") from None

        match resp_type:
            case cls.STATUS: return StatusResponse.from_dict(d)
            case cls.GET_KEY: return GetKeyResponse.from_dict(d)
            case cls.SUCCESS: return SuccessResponse.from_dict(d)
            case cls.FAILURE: return FailureResponse.from_dict(d)
            case cls.HANDSHAKE: return HandshakeResp.from_dict(d)

        raise ValueError(f"Got unexpected response type: {d.get("type")}") from None

class StatusResponse(SessionResp):
    RESP: ClassVar[str] = SessionResp.STATUS

    def __init__(self, remaining_ttl: int, user: str) -> None:
        self.remaining_ttl = remaining_ttl
        self.user = user

    def to_dict(self) -> Dict:
        return {
            "type": self.RESP,
            "ttl": self.remaining_ttl,
            "u": self.user,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(
            remaining_ttl=d["ttl"],
            user=d["u"],
        )


class GetKeyResponse(SessionResp):
    RESP: ClassVar[str] = SessionResp.GET_KEY

    def __init__(self, session_key: str) -> None:
        self.session_key = session_key

    def to_dict(self) -> Dict:
        return {
            "type": self.RESP,
            "sk": self.session_key,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["sk"])


class SuccessResponse(SessionResp):
    RESP: ClassVar[str] = SessionResp.SUCCESS

    def __init__(self, message: str) -> None:
        self.message = message

    def to_dict(self) -> Dict:
        return {
            "type": self.RESP,
            "msg": self.message,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["msg"])


class FailureResponse(SessionResp):
    RESP: ClassVar[str] = SessionResp.FAILURE

    def __init__(self, message: str) -> None:
        self.message = message

    def to_dict(self) -> Dict:
        return {
            "type": self.RESP,
            "msg": self.message,
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["msg"])

class HandshakeResp(SessionResp):
    RESP: ClassVar[str] = SessionResp.HANDSHAKE

    def __init__(self, all_good: bool) -> None:
        self.all_good = all_good

    def to_dict(self) -> Dict:
        return {
            "type": self.RESP,
            "g": self.all_good
        }

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["g"])




from typing import Annotated
from typer import Typer, Argument

from rune.commands.logincmd import handle_login_command, handle_logout_command

def setup(app: Typer):
    @app.command(name="login")
    def login(
        username: Annotated[
            str,
            Argument(help="Set the active username (namespace root).")
        ],
    ):
        """
        Set the active username for rune.

        The username acts as the root namespace for all secrets.
        This does not perform authentication or unlock encryption.
        """
        handle_login_command(username)


    @app.command(name="logout")
    def logout():
        """
        Clear the active username.

        After logout, no secrets can be accessed until a user is selected again.
        """
        handle_logout_command() 



from rune.models.settings.settings import Settings
from rune.models.settings.storagesettings import FileBasedStorageSettings
from rune.storage.secrets.base import StorageManager
from rune.storage.secrets.local import LocalJsonStorageManager
from pathlib import Path

def get_storage_manager(settings: Settings) -> StorageManager:
    match settings.storage:
        case FileBasedStorageSettings(file=file):
            return LocalJsonStorageManager(Path(file).absolute())
        case _:
            raise ValueError(f"Storage manager does not exist for identifier {settings.storage.mode}.")





from rune.crypto.base import Encrypter
from rune.crypto.aesgcm import AESGCMEncrypter
from rune.models.settings.encryptionsettings import AES_GCMEncryptionSettings
from rune.models.settings.settings import Settings

def get_encrypter(settings: Settings) -> Encrypter:
    match settings.encryption:
        case AES_GCMEncryptionSettings():
            return AESGCMEncrypter()
        case _:
            raise ValueError(f"Algorithm '{settings.encryption.mode}' is not supported.")

def get_encrypter_by_algorithm(algorithm: str | None) -> Encrypter:
    if algorithm == AESGCMEncrypter.encryption_algorithm():
        return AESGCMEncrypter()

    raise ValueError(f"Algorithm '{algorithm}' is not supported.")





from typer import Typer

from rune.utils import display
from rune.cli import configcli, secretcli, usercli, sessioncli, profilecli
from rune.context import Context
from rune.exception.badinput import BadInputError

def main():
    app = Typer(context_settings={"help_option_names": ["-h", "--help"]})

    secretcli.setup(app)
    usercli.setup(app)

    config_app = configcli.setup()
    config_app.add_typer(profilecli.setup())

    app.add_typer(sessioncli.setup())
    app.add_typer(config_app)

    try:
        Context.build()
        app()
    except BadInputError as e:
        display.failed_panel(e.message, title="[red]Bad input[/]")
    finally:
        shutdown()

def shutdown():
    Context.get().save_settings()





from rune.models.settings.sessionsettings import DaemonSessionSettings, SessionSettings
from rune.models.settings.settings import Settings
from rune.session.base import SessionManager
from rune.session.daemonmgr import DaemonSessionManager


def get_session_manager(settings: Settings) -> SessionManager:
    session_settings = settings.session
    match session_settings.mode:
        case SessionSettings.DAEMON if isinstance(session_settings, DaemonSessionSettings):
            return DaemonSessionManager("localhost", session_settings.port)
        case _:
            raise ValueError(f"Unexpected session mode: {session_settings.mode}")




from typing import Dict
from rich.console import Console
from rich.panel import Panel
from rich.pretty import Pretty
from rich.tree import Tree


console = Console()

def panel(message: str, title: str | None = None) -> None:
    console.print(Panel.fit(message, title=title))

def success_panel(message: str, title: str = "Success"):
    console.print(Panel.fit(message,title=f"[green]{title}[/]"))

def failed_panel(message: str, title: str = "Failed"):
    console.print(Panel.fit(message,title=f"[red]{title}[/]"))

def print(message: str):
    console.print(message)

def print_tree(tree: Tree):
    console.print(tree)

def print_dict(d: Dict):
    console.print(Panel.fit(Pretty(d, expand_all=True, indent_guides=True)))




from rune.crypto.base import Encrypter
from rune.exception.wrongencryption import WrongEncryptionMode
from rune.exception.wrongkey import WrongKeyUsed
from rune.models.crypto.secretfield import SecretField

import os
import base64
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

class AESGCMEncrypter(Encrypter):

    @classmethod
    def encryption_algorithm(cls) -> str:
        return "aesgcm"

    def __init__(self) -> None:
        self._encryption_algorithm = self.encryption_algorithm()

    def derive_key(self, password: str, salt: bytes) -> bytes:
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=600000
        )
        return kdf.derive(password.encode())


    def encrypt(self, secret: str, key: str, **kwargs) -> SecretField:
        """
        Encrypt a secret using a password-based key.
        Returns a dictionary suitable for JSON storage.
        """
        salt = os.urandom(16)
        nonce = os.urandom(12)

        aes_key = self.derive_key(key, salt)
        aesgcm = AESGCM(aes_key)

        ciphertext = aesgcm.encrypt(nonce, secret.encode(), None)

        return SecretField(
            ciphertext=base64.b64encode(ciphertext).decode("utf-8"),
            salt=base64.b64encode(salt).decode("utf-8"),
            nonce=base64.b64encode(nonce).decode("utf-8"),
            algorithm=self._encryption_algorithm
        )

    def decrypt(self, secret: SecretField, key: str, **kwargs) -> str:
        """
        Decrypt a secret previously encrypted by encrypt().
        """
        if self._encryption_algorithm != secret.algorithm:
            raise WrongEncryptionMode(f"Secret was encrypted with mode {secret.algorithm}. Please use it to decrypt.")

        salt = base64.b64decode(secret.salt or "")
        nonce = base64.b64decode(secret.nonce or "")
        ciphertext = base64.b64decode(secret.ciphertext)

        aes_key = self.derive_key(key, salt)
        aesgcm = AESGCM(aes_key)

        try:
            plaintext = aesgcm.decrypt(nonce, ciphertext, None)
            return plaintext.decode()
        except:
            raise WrongKeyUsed("Invalid key or corrupted secret")




from typer import Argument, Typer, Option
from typing import Annotated
import typer

from rune.context import Context
from rune.utils.input import get_choice_by_idx
from rune.utils import display

def setup() -> Typer:
    profile_app = Typer(name="profile", help="Manage configuration profiles. `rune config profile -h` for more help.")

    @profile_app.command(name="list")
    def list_profiles(
        interactive: Annotated[
            bool,
            Option(
                "--interactive",
                "-i",
                help="Interactively select a profile to activate after listing.",
            ),
        ] = False
    ):
        """
        List all configured profiles.
        """
        settings_manager = Context.get().settings_manager

        profiles = list(settings_manager.get_profiles().keys())
        profiles_file = str(settings_manager.profiles_file.absolute())

        display.print(f"[bold]Profiles file:[/] [cyan]{profiles_file}[/]")

        if not profiles:
            display.panel("[yellow]No profiles configured yet.[/] Create one with:\n[bold cyan]rune config profile save <profile-name>[/]")
            return

        display.success_panel("\n".join(f"[bold cyan][{idx}][/] {profile}" for idx, profile in enumerate(profiles, 1)), title="Available Profiles")

        if not interactive:
            return

        choice = get_choice_by_idx("Select profile to activate", profiles)
        if not choice:
            return

        load_profile(choice)

    @profile_app.command(name="save")
    def save_profile(
        _name: Annotated[
        str, Argument(help="Name under which the current configuration will be saved.")
    ],
        _force: Annotated[
        bool,
        Option(
            "--force",
            "-f",
            help="Overwrite the profile if it already exists.",
        ),
    ] = False,
    ):
        """
        Save the current configuration as a profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        if _name in settings_manager.get_profiles() and not _force:
            display.failed_panel(f"Profile '[bold cyan]{_name}[/]' already exists.\n\nUse [bold cyan]--force[/] to overwrite it.")
            return

        settings_manager.save_profile(context.settings, _name)
        display.success_panel(f"Profile '[bold cyan]{_name}[/]' saved successfully.")

    @profile_app.command(name="load")
    def load_profile(
        _name: Annotated[
        str, Argument(help="Name of the profile to activate.")
    ],
    ):
        """
        Activate a saved configuration profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        settings = settings_manager.get_profile(_name)
        if not settings:
            display.failed_panel(f"Profile '[bold cyan]{_name}[/]' does not exist.")
            return

        context.settings = settings.dirty()
        settings_manager.save_profile(settings, _name)

        display.success_panel(f"Switched to profile '[bold cyan]{_name}[/]'.")

    @profile_app.command(name="delete")
    def delete_profile(
        _name: Annotated[
        str, Argument(help="Name of the profile to delete.")
    ],
    ):
        """
        Delete an existing profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        profile = settings_manager.get_profile(_name)
        if not profile:
            display.failed_panel(f"Profile '[bold cyan]{_name}[/]' does not exist.")
            return

        if typer.confirm(f"Are you sure you want to delete profile '{_name}'?"):
            settings_manager.delete_profile(_name)
            display.success_panel(f"Profile '[bold cyan]{_name}[/]' deleted.")

        else:
            raise typer.Abort()

    return profile_app




from rune.internal.delete import delete_secret, delete_secret_fields, restore_secret
from rune.utils.input import get_fields_keys, get_session_key, input_key, input_name, sanitize_name
from rune.utils import display

def handle_delete_command(user: str, hard: bool, _name: str | None, _key: str | None, use_session_key: bool) -> None:
    name = sanitize_name(_name or input_name())

    session_key = get_session_key(user) if use_session_key else None
    key = _key or session_key or input_key()

    result = delete_secret(user, name, hard, key)

    if result.is_success():
        display.success_panel(f"[bold green]Deleted secret[/] [cyan]{name}[/]")
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")

def handle_delete_fields_command(user: str, hard: bool, _name: str | None, _key: str | None, _fields: str, use_session_key: bool) -> None:
    name = sanitize_name(_name or input_name())
    fields = get_fields_keys(_fields)

    session_key = get_session_key(user) if use_session_key else None
    key = _key or session_key or input_key()

    result = delete_secret_fields(user, name, hard, key, fields)

    if result.is_success():
        display.success_panel(f"[bold green]Deleted secret fields {", ".join(fields)}[/] for secret [cyan]{name}[/]")
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")


def handle_restore_cmd(user: str, _name: str | None) -> None:
    name = sanitize_name(_name or input_name())

    result = restore_secret(user, name)

    if result.is_success():
        display.success_panel(f"[bold green]Restored secret[/] [cyan]{name}[/]")
        return
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")




class NoSessionError(RuntimeError):
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message

class WrongUserError(RuntimeError):
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message




from typing import Annotated, Optional
import typer
from typer import Typer

from rune.commands.addcmd import handle_add_cmd
from rune.commands.getcmd import handle_get_command
from rune.commands.movecmd import handle_move_command
from rune.commands.updatecmd import handle_update_command
from rune.commands.deletecmd import handle_delete_command, handle_delete_fields_command, handle_restore_cmd
from rune.commands.listcmd import handle_ls_command
from rune.utils.input import ensure_active_user

NAME_HELP = (
    "The name of the secret. Supports namespaces (e.g., `db/prod/my-db`). "
    "If omitted, you will be prompted."
)

FIELDS_HELP = (
    "Comma-separated fields to store. You can provide `key=value` pairs "
    "for known values. Fields without a value will be queried securely.\n"
    "If omitted, will store a single-field secret where the field has the same name as the secret.\n"
    "Example: -f host=localhost,port=9999,user,password"
)

KEY_HELP = (
    "Encryption key to use. If omitted, the active session key will be used (if available).\n"
    "If no active session key is found, it will be queried securely."
)

KEY_HELP_UPDATE = (
    "Encryption key to use. Should be the same as the one used to originally encrypt the secret\n"
    "If omitted, the active session key will be used (if available).\n"
    "If no active session key is found, it will be queried securely."
)

def setup(app: Typer):

    @app.command()
    def add(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _fields: Annotated[Optional[str], typer.Option("--fields", "-f", help=FIELDS_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP)] = None,
        no_session_key: Annotated[bool, typer.Option("--no-session-key", help="Use --no-session-key to force encryption key input.")] = False,
    ):
        """
        Add a new secret to the vault.

        Supports namespaced secrets. Fields without explicit values
        will be prompted securely.
        """
        active_user = ensure_active_user()
        handle_add_cmd(active_user, _fields, _name, _key, not no_session_key)

    @app.command()
    def delete(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _fields: Annotated[
            Optional[str], 
            typer.Option("--fields", "-f", help="The fields to delete. Will delete all fields if not provided. Ex: -f <field1>,<field2>")
        ] = None,
        _hard: Annotated[bool, typer.Option("--hard", help="Hard delete the secret. Requires encryption key.")] = False,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help="Key used to encrypt the secret. Required if `--hard`.")] = None,
        no_session_key: Annotated[bool, typer.Option("--no-session-key", help="Use --no-session-key to force encryption key input.")] = False,
    ):
        """
        Delete a secret from the vault.

        Hard delete requires original encryption key.
        """
        active_user = ensure_active_user()
        if not _fields:
            handle_delete_command(active_user, _hard, _name, _key, not no_session_key)
        else:
            handle_delete_fields_command(active_user, _hard, _name, _key, _fields, not no_session_key)

    @app.command()
    def restore(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None
    ):
        """
        Restore a soft deleted secret.
        """
        active_user = ensure_active_user()
        handle_restore_cmd(active_user, _name)


    @app.command()
    def update(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _fields: Annotated[Optional[str], typer.Option("--fields", "-f", help=FIELDS_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP_UPDATE)] = None,
        no_session_key: Annotated[bool, typer.Option("--no-session-key", help="Use --no-session-key to force encryption key input.")] = False,
    ):
        """
        Update an existing secret in the vault.

        Fields without explicit values will be prompted securely.
        """
        active_user = ensure_active_user()
        handle_update_command(active_user, _fields, _name, _key, not no_session_key)

    @app.command()
    def get(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP)] = None,
        show: Annotated[bool, typer.Option("--show","-s",help="Show secret values in the terminal instead of hiding them.")]=False,
        show_deleted: Annotated[bool, typer.Option("--show-deleted", help="Show soft deleted fields")]=False,
        no_session_key: Annotated[bool, typer.Option("--no-session-key", help="Use --no-session-key to force encryption key input.")] = False,
    ):
        """
        Retrieve a secret from the vault.

        Copies the selected field to clipboard by default.
        """
        active_user = ensure_active_user()
        handle_get_command(active_user, _name, _key, show, show_deleted, not no_session_key)

    OG_NAME_HELP = "Full name of secret to move"
    DEST_NAME_HELP = "Destination name for the secret"
    @app.command(name="move")
    def move(
        _original_name: Annotated[Optional[str], typer.Argument(help=OG_NAME_HELP)] = None,
        _new_name: Annotated[Optional[str], typer.Argument(help=DEST_NAME_HELP)] = None
    ):
        """
        Move a secret from one name to another.
        """
        active_user = ensure_active_user()
        handle_move_command(active_user, _original_name, _new_name)

    @app.command(name="ls")
    def list_entries(
        namespace: Annotated[Optional[str], typer.Argument(help="Filter secrets by namespace")] = None,
        interactive: Annotated[bool, typer.Option("--interactive", "-i", help="Interactively select and retrieve secrets from the list.")] = False,
        show: Annotated[bool, typer.Option("--show","-s", help="Show secret values in the terminal. Only used with --interactive.")]=False,
        show_deleted: Annotated[bool, typer.Option("--show-deleted", help="Show soft deleted secrets.")] = False,
        no_session_key: Annotated[bool, typer.Option("--no-session-key", help="Use --no-session-key to force encryption key input.")] = False,
    ):
        """
        List all secrets in the vault for the logged in user, organized by namespace.

        Single-child namespaces are collapsed for cleaner display.
        Use `--namespace` to filter results.
        """
        active_user = ensure_active_user()
        handle_ls_command(active_user, namespace, interactive, show, show_deleted, not no_session_key)

